/*
 * Copyright (C) 2020 Christian Zietz
 * Parts: Copyright (C) 2001-2019 The EmuTOS development team
 *
 * This file is distributed under the GPL, version 2 or at your
 * option any later version.  See doc/license.txt for details.
 */

#include <osbind.h>

#include "rtc_driver.h"

static UBYTE int2bcd(UWORD a)
{
    return (a % 10) + ((a / 10) << 4);
}

static UWORD bcd2int(UBYTE a)
{
    return (a & 0xf) + ((a >> 4) * 10);
}

#define MONSTER_I2C_DIR *(volatile unsigned short *)0xfffffe02L
#define MONSTER_I2C_SCL *(volatile unsigned short *)0xfffffe04L
#define MONSTER_I2C_SDA *(volatile unsigned short *)0xfffffe06L

#define I2C_HIGH 1 
#define I2C_LOW 0

/* More than plenty even on an accelerated 68000. */
#define DELAY5US { int k; for (k=0; k<10; k++) __asm volatile("nop"); }

static void i2c_start (void)
{
    MONSTER_I2C_SDA = I2C_HIGH;
    DELAY5US;
    MONSTER_I2C_SCL = I2C_HIGH;
    DELAY5US;
    MONSTER_I2C_SDA = I2C_LOW;
    DELAY5US;
    MONSTER_I2C_SCL = I2C_LOW;
}

static void i2c_stop (void)
{
    MONSTER_I2C_SCL = I2C_LOW;
    MONSTER_I2C_SDA = I2C_LOW;
    MONSTER_I2C_SCL = I2C_HIGH;
    DELAY5US;
    MONSTER_I2C_SDA = I2C_HIGH;

    MONSTER_I2C_DIR = 3; /* Set pins to float */
}

static void i2c_write (UBYTE data)
{
    UBYTE i;

    for(i = 0; i < 8; i++)
    {
        MONSTER_I2C_SDA = (data & 0x80) ? 1 : 0;
        data <<= 1;
        MONSTER_I2C_SCL = I2C_HIGH;
        DELAY5US;
        MONSTER_I2C_SCL = I2C_LOW;
    }

    MONSTER_I2C_DIR = 2; /* SDA as input */
    MONSTER_I2C_SCL = I2C_HIGH;
    DELAY5US;
    MONSTER_I2C_SCL = I2C_LOW;
}

static UBYTE i2c_read (void)
{
    UBYTE i, data = 0;

    MONSTER_I2C_DIR = 2; /* SDA as input */

    for(i = 0; i < 8; i++)
    {
        data <<= 1;
        MONSTER_I2C_SCL = I2C_HIGH;
        data |= MONSTER_I2C_SDA;
        DELAY5US;
        MONSTER_I2C_SCL = I2C_LOW;
    }

    return data;
}

void write_ds1307(UBYTE address, UBYTE data)
{
    i2c_start();
    i2c_write(0xd0);
    i2c_write(address);
    i2c_write(data);
    i2c_stop();
}

UBYTE read_ds1307(UBYTE address)
{
    UBYTE data;

    i2c_start();
    i2c_write(0xd0);
    i2c_write(address);
    i2c_start();
    i2c_write(0xd1);
    data = i2c_read();
    i2c_stop();

    return data;
}

static void puts_twodigits(int x, const char* delim)
{
    int t,o;
    x = x % 100;
    t = x / 10;
    o = x % 10;
    Cconout(t+'0');
    Cconout(o+'0');
    if (delim)
        Cconws(delim);
}

ULONG rtc_gettime(BOOL print_it)
{
    int sec, min, hr, day, mon, year;

    /* Read from RTC via I2C. */
    sec  = bcd2int(read_ds1307(0));
    min  = bcd2int(read_ds1307(1));
    hr   = bcd2int(read_ds1307(2) & 0x3f); /* because MSBs encode 12/24h mode */
    
    day  = bcd2int(read_ds1307(4));
    mon  = bcd2int(read_ds1307(5) & 0x7f); /* because MSB encodes century */
    year = bcd2int(read_ds1307(6)) + 2000; /* always assume 21st century */
    
    /* Reading 0 as day means RTC not present or not initialized. */
    if (print_it && (day != 0))
    {
        Cconws("RTC Time: ");
        puts_twodigits(day, ".");
        puts_twodigits(mon, ".20");
        puts_twodigits(year, " ");
        puts_twodigits(hr, ":");
        puts_twodigits(min, ":");
        puts_twodigits(sec, "\r\n");
    }
    
    /* Encode into XBIOS format. */
    ULONG t = 0;
    t |= (ULONG)sec/2;
    t |= (ULONG)min << 5;
    t |= (ULONG)hr  << 11;
    t |= (ULONG)day << 16;
    t |= (ULONG)mon << 21;
    t |= (ULONG)(year-1980) << 25;
    
    return t;
}

void rtc_settime(ULONG time)
{
    int sec, min, hr, day, mon, year;
    
    /* Decode from XBIOS format. */
    hr   = (time >> 11) & 0x1f;
    min  = (time >> 5) & 0x3f;
    sec  = (time << 1) & 0x3f;
    time >>= 16;
    year = (time >> 9) - 20;
    mon  = (time >> 5) & 0x0f;
    day  = time & 0x1f;
    
    /* Write to RTC via I2C. */
    write_ds1307(0, int2bcd(sec));
    write_ds1307(1, int2bcd(min));
    write_ds1307(2, int2bcd(hr));
    
    write_ds1307(4, int2bcd(day));
    write_ds1307(5, int2bcd(mon));
    write_ds1307(6, int2bcd(year));
}

/* Default date for uninitialized RTC 01.01.2020, 00:00:00 */
#define DEFAULT_DATE ((1ul<<16) | (1ul<<21) | (40ul<<25))

/* Must be run in supervisor mode */
LONG rtc_detect(void)
{
    LONG err = 0;

    /* check for register accessibility */
    __asm volatile(
        "moveq.l    #-1,%0\n\t"     // assume failure
        "move.l     8.w,a1\n\t"     // save old bus error vector
        "move.l     #buserr,8.w\n\t"// new bus error vector
        "move.l     sp,a0\n\t"      // save stack pointer
        "tst.w      0xfffffe06.w\n\t"  // I2C SDA register
        "nop\n\t"
        "moveq.l    #0,%0\n"         // success!
        "buserr:\n\t"
        "move.l     a1,8.w\n\t"     // restore bus error vector
        "move.l     a0,sp"          // restore stack pointer
        : "=d"(err) : : "a0","a1","cc");

    if (err)
        return err;

    /* read day register: if it's zero RTC is either uninitialized or not present */
    if (read_ds1307(4) == 0)
    {
        /* try to set default date. */
        rtc_settime(DEFAULT_DATE);
        if (read_ds1307(4) == 0) {
            /* RTC not present */
            err = -2;
        }
    }

    return err;
}
