/*
 * Copyright (C) 2020 Christian Zietz
 *
 * This file is distributed under the GPL, version 2 or at your
 * option any later version.  See doc/license.txt for details.
 */

#include <string.h>
#include <osbind.h>
#include <gem.h>

#include "rtc_set.rsh"
#include "rtc_driver.h"

static void fill_twodigits(char *y, int x)
{
    int t,o;
    x = x % 100;
    t = x / 10;
    o = x % 10;
    y[0] = t+'0';
    y[1] = o+'0';
}

static void fill_date_time(char *datestr, char *timestr, unsigned long time)
{
    int sec, min, hr, day, mon, year;

    /* Decode from XBIOS format. */
    hr   = (time >> 11) & 0x1f;
    min  = (time >> 5) & 0x3f;
    sec  = (time << 1) & 0x3f;
    time >>= 16;
    year = (time >> 9) - 20;
    mon  = (time >> 5) & 0x0f;
    day  = time & 0x1f;
    
    fill_twodigits(&datestr[0], day);
    fill_twodigits(&datestr[2], mon);
    fill_twodigits(&datestr[4], year);
    
    fill_twodigits(&timestr[0], hr);
    fill_twodigits(&timestr[2], min);
    fill_twodigits(&timestr[4], sec);
}

static int parse_twodigits(char *y)
{
    return (y[0]-'0')*10 + (y[1]-'0');
}

static unsigned long parse_date_time(char *datestr, char *timestr)
{                             // J   F   M   A   M   J   J   A   S   O   N   D
    char days_per_month[] = {0, 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31};
    int sec, min, hr, day, mon, year;
    
    /* parse */
    day  = parse_twodigits(&datestr[0]);
    mon  = parse_twodigits(&datestr[2]);
    year = parse_twodigits(&datestr[4]);
    
    hr  = parse_twodigits(&timestr[0]);
    min = parse_twodigits(&timestr[2]);
    sec = parse_twodigits(&timestr[4]);
    
    /* validate */
    if ((hr < 0) || (hr > 23))
        return 0;
    if ((min < 0) || (min > 59))
        return 0;
    if ((sec < 0) || (sec > 59))
        return 0;

    if ((year < 0) || (year > 99))
        return 0;
    if ((mon < 1) || (mon > 12))
        return 0;
    
    /* leap year? correct for 2000 - 2099. */
    if ((year % 4) == 0)
        days_per_month[2] = 29;

    if ((day < 1) || (day > days_per_month[mon]))
        return 0;

    /* Encode into XBIOS format */
    unsigned long t = 0;
    t |= (ULONG)sec/2;
    t |= (ULONG)min << 5;
    t |= (ULONG)hr  << 11;
    t |= (ULONG)day << 16;
    t |= (ULONG)mon << 21;
    t |= (ULONG)(year+20) << 25;
    
    return t;
}

int main(void)
{
    short i, dx, dy, dw, dh, exit, cont = 1;
    unsigned long time;

    appl_init();

    for(i = 0; i < NUM_OBS; i++)
        rsrc_obfix(rs_object, i);

    graf_mouse(0, 0L);

#ifndef USE_XBIOS
    if (Supexec(rtc_detect) != 0)
    {
        form_alert(1, "[3][No RTC detected!][EXIT]");
        appl_exit();
        return 1;
    }

    time = rtc_gettime(FALSE);

#else
    time = Gettime(); /* from XBIOS */
#endif

    fill_date_time(rs_object[ED_DATE].ob_spec.tedinfo->te_ptext,
                   rs_object[ED_TIME].ob_spec.tedinfo->te_ptext,
                   time);

    form_center(rs_object, &dx, &dy, &dw, &dh);
    form_dial(FMD_START, dx,dy,0,0, dx,dy,dw,dh);

    while (cont)
    {
        objc_draw(rs_object, 0, 8, dx, dy, dw, dh);
        exit = form_do(rs_object, 0);
        switch (exit) {
            case BT_SET:
                time = parse_date_time(rs_object[ED_DATE].ob_spec.tedinfo->te_ptext,
                                       rs_object[ED_TIME].ob_spec.tedinfo->te_ptext);
                if (time != 0) {
#ifndef USE_XBIOS
                    rtc_settime(time);
#endif
                    /* also set GEMDOS date/time */
                    Tsetdate(time >> 16);
                    Tsettime(time & 0xFFFFu);
                    cont = 0;
                } else {
                    form_alert(1, "[1][Invalid date or time!][ OK ]");
                    rs_object[BT_SET].ob_state &= ~OS_SELECTED; /* reset button selection */
                }
                break;
            case BT_CANCEL:
                cont = 0;
                break;
            default:
                break;
        }
    }
    
    form_dial(FMD_FINISH, 0,0,0,0,0,0,0,0);
    appl_exit();
    return 0;
}